/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Mnenhy.
 *
 * The Initial Developer of the Original Code is Karsten Duesterloh.
 *
 * Portions created by the Initial Developer are Copyright (C) 2003
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *  Karsten Duesterloh <mnenhy@tprac.de>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */
// +-
// |    mnenhy-msgHdrViewOverlay.js: JavaScripts for MailNews enhancements
// |
// |    Written, forged, copied and conceived by Karsten Duesterloh <mnenhy@tprac.de>
// +-

// minimum usable version of messageidfinder
const gksMIDFVersionCurrent = InstallTrigger.getVersion("messageidfinder");
const gksMIDFVersionMinimum = "1.8";
const gksMIDF               = "messageidfinder-menu";
// special newsgroup value
const gksPoster             = "poster";


// need to remember last header view mode to avoid flashing
// this causes the headers to be shown not necessarily in their real order!
var giHeaderView = cViewBriefDirty;

function initializeHeaderViewTables()
{
  var viewmode = GetHeaderView();
  if (viewmode == cViewNoHeaders)
  { // delete views
    gCollapsedHeaderView = {};
    gExpandedHeaderView  = {};
  }
  else
  { // Build HeaderViews

    // Remarks: * if (!gXXXHeaderView.length) is useless, because gXXXHeaderView is associative!
    //          * isEmailAddress is used nowhere and hence ignored here

    // CollapsedHeaderView
    for (var index = 0; index < gCollapsedHeaderList.length; index++)
    {
      gCollapsedHeaderView[gCollapsedHeaderList[index].name] =
        new createHeaderEntry('collapsed', gCollapsedHeaderList[index]);
    }

    // ExpandedHeaderView
    if (viewmode >= cViewCustomBase)
    { // create new gExpandedHeaderList from custom headers in mnenhy.rdf
      const ksToggle = "useToggle";
      const ksKnown  = "header:known";
      const ksOutput = "format";
      const kfOutput = [updateHeaderValue,
                        OutputEmailAddresses,
                        MnenhyOutputNewsgroups,
                        MnenhyOutputReferences];

      gExpandedHeaderList = [];
      var asHeaderList = gMnenhyRDF.GetNodeList("header:custom" + (viewmode - cViewCustomBase));
      for (var iHeader in asHeaderList)
      {
        var sHeader = asHeaderList[iHeader];
        var bToggle = !!   gMnenhyRDF.GetItemValue(ksToggle, sHeader, ksKnown);
        var iOutput = eval(gMnenhyRDF.GetItemValue(ksOutput, sHeader, ksKnown));
        var fOutput = kfOutput[(iOutput in kfOutput) ? iOutput : gkiOutputText];
        gExpandedHeaderList[iHeader] = {name:          sHeader,
                                        useToggle:     bToggle,
                                        outputFunction:fOutput};
      } // for (iHeader in asHeaderList)
    } // if custom view
    else
    { // old default HeaderList from msgHdrViewOverlay
      // (isEmailAddress was never used!)
      gExpandedHeaderList = [ {name:"subject"},
                              {name:"from",        outputFunction:OutputEmailAddresses},
                              {name:"reply-to",    outputFunction:OutputEmailAddresses},
                              {name:"date"},
                              {name:"to",          outputFunction:OutputEmailAddresses, useToggle:true},
                              {name:"cc",          outputFunction:OutputEmailAddresses, useToggle:true},
                              {name:"bcc",         outputFunction:OutputEmailAddresses, useToggle:true},
                              {name:"newsgroups",  outputFunction:OutputNewsgroups},
                              {name:"followup-to", outputFunction:OutputNewsgroups} ];
    }

    // sync current gExpandedHeaderView and gExpandedHeaderList
    for (index = 0; index < gExpandedHeaderList.length; index++)
    {
      var headerName = gExpandedHeaderList[index].name;
      gExpandedHeaderView[headerName] = new MnenhyCreateExpandedView(gExpandedHeaderList[index]);
    }
  } // Build HeaderViews
}


function MnenhyOutputReferences(headerEntry, headerValue)
{ // do it like NS 4.x did: show linkified numbers instead of actual MID
  // <MID> => [1] -> news://<currentServer>/MID
  // this Link should be copyable and clicking should call the MIDF, if present
  headerEntry.enclosingBox.clearEmailAddresses();
  var rsRefs = headerValue.match(/[^<>\s]+/g);  // MIDs without <>
  for (var iRef = 0; iRef < rsRefs.length; ++iRef)
  {
    var oRef = {sMID:   rsRefs[iRef],
                iIndex: iRef + 1};
    headerEntry.enclosingBox.addAddressView(oRef);
  }
  headerEntry.enclosingBox.buildViews(-1);  // show all
}


function MnenhyOutputNewsgroups(headerEntry, headerValue)
{ // do it like NS 4.x did: show linkified newsgroups
  // <NG> => [NG] -> news://<currentServer>/NG
  // this Link should be copyable and clicking should call the MIDF, if present
  headerEntry.enclosingBox.clearEmailAddresses()
  var rsNGs = headerValue.split(",");
  for (var iNG = 0; iNG < rsNGs.length; ++iNG)
  {
    headerEntry.enclosingBox.addAddressView(rsNGs[iNG]);
  }
  headerEntry.enclosingBox.buildViews(gNumAddressesToShow);  // usually show only 3
}


function UpdateMessageHeaders()
{
  // iterate over each header we received and see if we have a matching entry
  // in each header view table...
  for (var headerName in currentHeaderData)
  {
    var headerField  = currentHeaderData[headerName];
    var sDisplayName = headerField.headerName;
    var headerEntry  = null;

    if (gCollapsedHeaderViewMode && !gBuiltCollapsedView)
    {
      headerEntry = gCollapsedHeaderView[headerName];
      if (headerEntry != undefined && headerEntry)
      {
        headerEntry.outputFunction(headerEntry, headerField.headerValue);
        headerEntry.valid = true;
      }
    }
    else if (!gCollapsedHeaderViewMode && !gBuiltExpandedView)
    {
      if (headerName in gExpandedHeaderView)
      { // use existing view
        headerEntry = gExpandedHeaderView[headerName];
      }
      else
      { // create new view
        sDisplayName = sDisplayName.toLowerCase();
        
        var viewmode = GetHeaderView();
        if (viewmode >= cViewCustomBase)
        { // for custom view, also check the display name
          if (sDisplayName in gExpandedHeaderView)
          { // create new view item
            headerEntry = new MnenhyCreateExpandedView({name: headerName})
            gExpandedHeaderView[headerName] = headerEntry;
          }
        }
        else
        { // non-custom view modes
          switch (viewmode)
          {
            case cViewAllHeaders:
              { // for view all headers, create missing views
                currentHeaderData[headerName] = {headerName: Capitalize(sDisplayName)};
                headerEntry = new MnenhyCreateExpandedView({name: headerName})
                gExpandedHeaderView[headerName] = headerEntry;
              }
              break;

            case cViewBriefHeaders:
            case cViewNoHeaders:
            default:
              break;
          }
        } // non-custom view modes
      } // create new view

      if (headerEntry)
      {
        headerEntry.outputFunction(headerEntry, headerField.headerValue);
        headerEntry.valid = true;
      }
    } // if we are in expanded view....
  }

  if (gCollapsedHeaderViewMode)
      gBuiltCollapsedView = true;
  else
      gBuiltExpandedView = true;

  // now update the view to make sure the right elements are visible
  updateHeaderViews();

  if ("FinishEmailProcessing" in this)
      FinishEmailProcessing();
}


// modify the messageHeaderSink's onStartHeaders function
// if we don't, the header view does not get updated when we change from all to custom
messageHeaderSink.onStartHeaders = function ()
{
  var oElem = document.getElementById("msgHeaderView");
  if (GetHeaderView() == cViewNoHeaders)
  { // <sing>no headers, no box...</sing> ;-)
    oElem.setAttribute("collapsed","true");
  }
  else
  {
    oElem.setAttribute("collapsed","false");

    // clear out any pending collected address timers...
    if (gCollectAddressTimer)
    {
        gCollectAddress = "";
        clearTimeout(gCollectAddressTimer);
        gCollectAddressTimer = null;
    }

    var iHeaderView = GetHeaderView();
    if (iHeaderView != giHeaderView)
    {   // rebuild *complete* header view to sustain the headers' order
        hideHeaderView(gExpandedHeaderView);
        gExpandedHeaderView = {};
        initializeHeaderViewTables();
        giHeaderView = iHeaderView;
    }

    // set gViewAllHeaders (for backwards compatibility)
    gViewAllHeaders = (iHeaderView == cViewAllHeaders);


    ClearCurrentHeaders();
    gBuiltExpandedView = false;
    gBuiltCollapsedView = false;
    gBuildAttachmentsForCurrentMsg = false;
    gBuildAttachmentPopupForCurrentMsg = true;
    ClearAttachmentList();
    ClearEditMessageButton();

    for (var index in gMessageListeners)
        gMessageListeners[index].onStartHeaders();
  }
}


// modify the messageHeaderSink's processHeaders function
// if there're multiple headers of the same kind, we usually show them all.
// but in case of to, cc and bcc, we merge
// !!! the bug fixed here is fixed in the Mozilla trunk since 2002-12-10 (->1.3a) !
messageHeaderSink.processHeaders = function(headerNames, headerValues, numHeaders, dontCollectAddress)
{
    this.onStartHeaders();

    var index = 0;
    // process each header
    while (index < numHeaders)
    {
        // for consistancy sake, let's force all header names to be lower case so
        // we don't have to worry about looking for: Cc and CC, etc.
        var lowerCaseHeaderName = headerNames[index].toLowerCase();

        var foo = new Object;
        foo.headerValue = headerValues[index];
        foo.headerName = headerNames[index];

        var sUniqueName = lowerCaseHeaderName;
        var bOldHeader = (sUniqueName in currentHeaderData);
        if (bOldHeader)
        {   // need to create an unique name
            // special handling for certain headers
            switch (lowerCaseHeaderName)
            {
                case "to":
                case "cc":
                case "bcc":
                    currentHeaderData[lowerCaseHeaderName].headerValue = currentHeaderData[lowerCaseHeaderName].headerValue + ',' + foo.headerValue;
                    break;

                default:
                    {
                        var iUID = 0;
                        do
                        {
                            sUniqueName = lowerCaseHeaderName + iUID;
                            ++iUID;
                        }
                        while (sUniqueName in currentHeaderData);
                        bOldHeader = false; // create this new header
                    }
                    break;
            }
        }

        if (!bOldHeader)
        {
            currentHeaderData[sUniqueName] = foo;
        }

        if (lowerCaseHeaderName == "from")
        {
            if (foo.headerValue && abAddressCollector)
            {
                if (  (gCollectIncoming  && !dontCollectAddress)
                    ||(gCollectNewsgroup &&  dontCollectAddress)
                   )
                {
                    gCollectAddress = foo.headerValue;
                    // collect, and add card if doesn't exist
                    gCollectAddressTimer = setTimeout('abAddressCollector.collectUnicodeAddress(gCollectAddress, true);', 2000);
                }
                else if (gCollectOutgoing)
                {
                    // collect, but only update existing cards
                    gCollectAddress = foo.headerValue;
                    gCollectAddressTimer = setTimeout('abAddressCollector.collectUnicodeAddress(gCollectAddress, false);', 2000);
                }
            }
        }
        index++;
    }

    this.onEndHeaders();
}


// if messageidfinder exists and has the right version, use it
var gbMIDF = null;    // null = not checked, true = found, false = not found
var goMIDF = null;    // the MIDF menu object

function GetMIDF()
{
  if (gbMIDF === null)
  { // not checked yet
    gbMIDF = !!gksMIDFVersionCurrent;
    if (gbMIDF)
    { // a MIDF exists, so check version
      gbMIDF = (gksMIDFVersionCurrent >= gksMIDFVersionMinimum);
      if (gbMIDF)
      {
        goMIDF = document.getElementById(gksMIDF);
        gbMIDF = !!goMIDF;
      }
    }
  }
  return goMIDF;
}


function SearchContextForMID(oNode)
{
  return (oNode.id == gksMIDF) ? NodeFilter.FILTER_ACCEPT : NodeFilter.FILTER_REJECT;
}

function OnPopupShowingMIDF(event)
{
  var oMenuPopUp = event.originalTarget;
  if (oMenuPopUp.id != "messageidfinder-popup")
  {
    var oTreeWalker = document.createTreeWalker(oMenuPopUp,
                                                NodeFilter.SHOW_ELEMENT,
                                                SearchContextForMID,
                                                false);
    var bNoMIDF = true;
    // actually, nextNode() should just return null if there's no such node...
    try{bNoMIDF = !oTreeWalker.nextNode();}catch(ignored){}
    if (bNoMIDF)
    {
      // +- 
      // |  call MIDF
      AddNewsServers();
      // +-
      var oMIDFinderMenu = GetMIDF().cloneNode(true);
      oMenuPopUp.appendChild(document.createElement("menuseparator"));
      oMenuPopUp.appendChild(oMIDFinderMenu);
    }
  }
}


// prefix --> the name of the view (i.e. "collapsed", "expanded")
// headerListInfo --> entry from a header list.
function MnenhyCreateExpandedView(oHeaderEntry)
{
  if (!oHeaderEntry)
    return;
  
  var sHeaderName    = oHeaderEntry.name;
  var sPartialIDName = "expanded" + sHeaderName;
  this.enclosingBox  = document.getElementById(sPartialIDName + "Box");
  this.textNode      = document.getElementById(sPartialIDName + "Value");
  this.isValid       = false;

  // if we already have a view box, add our class,
  // else create a new box
  if (this.enclosingBox)
  { // class override
    MnenhyAddExpandedBoxFeatures(this.enclosingBox, oHeaderEntry)
  }
  else 
  { // new box
    this.enclosingBox = new MnenhyCreateExpandedBox(oHeaderEntry);
  }
  
  if ("useToggle" in oHeaderEntry)
  {
    this.useToggle = oHeaderEntry.useToggle;
    if (this.useToggle) // find the toggle icon in the document
    {
      this.textNode = this.enclosingBox.emailAddresses;
    }
  }
  else
   this.useToggle = false;

  if ("useShortView" in oHeaderEntry)
  {
    this.useShortView = oHeaderEntry.useShortView;
    this.enclosingBox.emailAddressNode = this.textNode;
  }
  else
    this.useShortView = false;

  if ("outputFunction" in oHeaderEntry)
    this.outputFunction = oHeaderEntry.outputFunction;
  else
    this.outputFunction = updateHeaderValue;
}


function MnenhyCreateExpandedBox(oHeaderEntry)
{
  if (!oHeaderEntry)
    return null;
  
  const ksBoxIdName  = "expanded" + oHeaderEntry.name + "Box";
  const ksPrettyName = Capitalize(oHeaderEntry.name);
  // create pretty display name 
  currentHeaderData[oHeaderEntry.name] = {headerName: ksPrettyName};
  
  // create element according to outputFunction
  var sOutputFunction = "";   // use default output
  if (oHeaderEntry.outputFunction)
    sOutputFunction = oHeaderEntry.outputFunction.name;

  const ksFieldText  = "mail-headerfield";
  const ksFieldMulti = "mail-multi-emailHeaderField";

  var sNewElement;
  switch (sOutputFunction)
  {
    case "MnenhyOutputReferences":
      sNewElement = ksFieldMulti;
      break;

    case "MnenhyOutputNewsgroups":
      sNewElement = ksFieldMulti;
      break;

    case "OutputEmailAddresses":
      sNewElement = ksFieldMulti;
      break;

    default:
      sNewElement = ksFieldText;
  }

  // create new view and set attributes
  var oNewView = document.createElement(sNewElement);
  oNewView.setAttribute("id",             ksBoxIdName);
  oNewView.setAttribute("label",          ksPrettyName + ": ");
  oNewView.setAttribute("keywordrelated", "true");
  oNewView.collapsed = true;

  // this new element needs to be inserted into the view...
  var oTopViewNode = document.getElementById("expandedHeaders");
  oTopViewNode.appendChild(oNewView);

  MnenhyAddExpandedBoxFeatures(oNewView, oHeaderEntry);

  return oNewView;
}


// set new features like class, context, etc. according to the output function
function MnenhyAddExpandedBoxFeatures(oBox, oHeaderEntry)
{
  if (oBox && oHeaderEntry)
  {
    var sOutputFunction = "";   // use default output
    if (oHeaderEntry.outputFunction)
      sOutputFunction = oHeaderEntry.outputFunction.name;

    const ksClassRef     = "MnenhyRefField";
    const ksClassNG      = "MnenhyNGField";
    const ksClassText    = "MnenhyTextField";
    const ksClassPersist = "MnenhyToggleState";

    // add new class
    var sFunctionClass, bMIDFhandler;
    switch (sOutputFunction)
    {
      case "MnenhyOutputReferences": 
        sFunctionClass = ksClassRef; 
        bMIDFhandler   = false;
        break;

      case "MnenhyOutputNewsgroups": 
        sFunctionClass = ksClassNG;
        bMIDFhandler   = false;
        break;

      case "OutputEmailAddresses":   
        sFunctionClass = ksClassPersist;         
        bMIDFhandler   = false;
        break;

      default:                       
        sFunctionClass = ksClassText;
        bMIDFhandler   = GetMIDF();
    }

    // add class
    var sNewClass = oBox.getAttribute("class")
    if (sFunctionClass && (String(sNewClass).indexOf(sFunctionClass) < 0))
    {
      sNewClass += " " + sFunctionClass;
      sNewClass = sNewClass.replace(/^\s+|\s+$/g, "");      // trim whitespace
      oBox.setAttribute("class", sNewClass);
    }

    // add MIDF, if reasonable and possible
    if (bMIDFhandler)
      oBox.addEventListener("popupshowing", OnPopupShowingMIDF, false);
  }
}

  
function OnMnenhyRefClick(event)
{
  if (!event.button)
  { // left click
    if (GetMIDF())
    { // correct MIDF version installed
      SetWaitCursor();      
      // +- 
      // |  call MIDF
      AddNewsServers();
      var oMsgHdr = OpenMessageByMessageID(event.originalTarget.getAttribute("sMID"));
      // +- 
      UnsetWaitCursor();
    
      if (!oMsgHdr)
      {
        const ksFailed = GetMnenhyStr("MnDialog_MID_failed");
        MnenhyAlertService(ksFailed);
      }  
    }
    else
    {
      MnenhyAlert(GetMnenhyStr("MnDialog_MIDF_version"), GetMnenhyStr("MnDialog_MIDF_title"))
    }
  }
}


function OnMnenhyNGClick(event)
{
  if (!event.button)
  {
    const ksNG = event.originalTarget.getAttribute("sNG");
    if (ksNG == gksPoster)
    { // if clicked NG is poster, reply to sender
      goDoCommand('cmd_replySender');
    }
    else
    {
      const koServer = GetLoadedMsgFolder().server;
      const ksURI    = koServer.serverURI + "/" + ksNG;

      SetWaitCursor();      
      try
      {
        SelectFolder(ksURI);
      }
      catch(ex)
      { // could not select folder, maybe 'cos it ain't there yet
        UnsetWaitCursor();      
        if (ksURI.substr(0,7) == "news://")
        { // only try if in news folder
          const koSubServ = koServer.QueryInterface(Components.interfaces.nsISubscribableServer);
          if (GetSubscribeConfirmation(ksNG))
          {
            SetWaitCursor();      
            try
            {
              koSubServ.subscribe(ksNG);
              koSubServ.commitSubscribeChanges();
              SelectFolder(ksURI);  // retry selecting folder
            }
            catch(ex)
            { 
              UnsetWaitCursor();      
              const ksFailed = GetMnenhyStr("MnDialog_NG_failed").replace("%ng", ksNG);
              MnenhyAlertService(ksFailed);
            } // no group found on server
          } // subscribing allowed 
        } // we are on a news server
      }
      UnsetWaitCursor();      
    }
  }
}


function GetSubscribeConfirmation(sNG)
{
  try
  {
    var oNewsBundle = Components.classes['@mozilla.org/intl/stringbundle;1']
                      .getService(Components.interfaces.nsIStringBundleService)
                      .createBundle("chrome://messenger/locale/news.properties");
    var s = oNewsBundle.GetStringFromName("autoSubscribeText").replace("%1$S", sNG);
    return MnenhyConfirmService(s);
  }
  catch(error)
  {
    return false;
  }
}
             

function PrettifyMID(sMID)
{
  return "<" + sMID + ">";
}


// CopyMID to the clipboard
function CopyMID(oMID, bAll)
{
  if (oMID)
  {
    var sMID = "";
    if (bAll)
    {
      var roMIDs = oMID.parentNode.parentNode.mAddresses;
      for (var i = 0; i < roMIDs.length; ++i)
        sMID += PrettifyMID(roMIDs[i].sMID) + " ";
      sMID = sMID.replace(/\s+$/, "");
    }
    else
    {
      sMID = PrettifyMID(oMID.getAttribute("sMID"));
    }    
    CopyStringToClipboard(sMID);
  }
}


// CopyNG to the clipboard
function CopyNG(oNG, bAll)
{
  if (oNG)
  {
    var sNG = "";
    if (bAll)
    {
      var roNGs = oNG.parentNode.parentNode.mAddresses;
      for (var i = 0; i < roNGs.length; ++i)
        sNG += roNGs[i] + ",";
      sNG = sNG.replace(/,+$/, "");
    }
    else
    {
      sNG = oNG.getAttribute("sNG");
    }    
    CopyStringToClipboard(sNG);
  }
}

