/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Mnenhy.
 *
 * The Initial Developer of the Original Code is Karsten Duesterloh.
 *
 * Portions created by the Initial Developer are Copyright (C) 2003
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *  Karsten Duesterloh <mnenhy@tprac.de>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */
// +-
// |    mnenhy-pref-chroman.js
// |
// |    Written, forged, copied and conceived by Karsten Duesterloh <mnenhy@tprac.de>
// +-
// |    History:
// |    
// |    2003-02-16      total rewrite because of hard disk crash :(
// +-


// |
// | constants
// |
const gciPackType        = 0;
const gciPackAction      = 1;
const gciPackURL         = 2;
const gciPackPath        = 3;

const gciPackTypeAll     = 0;
const gciPackTypeContent = 1;
const gciPackTypeLocale  = 2;
const gciPackTypeSkin    = 3;
const gcasPackTypesA     = ["","content","locale","skin"];
const gcasPackTypesB     = {"content":gciPackTypeContent,"locale":gciPackTypeLocale,"skin":gciPackTypeSkin};
const gcsPackInstall     = "install";
const gcsPackURL         = "url";
const gcsPackResPathExpr = "(jar:)?resource:/chrome/%PackageName%(\.jar!)?/%PackType%/%Locale%(.*)/";
const gcsWordDelimExpr   = "\\b";
const gcsCR              = "\r";
const gcsLF              = "\n";
const gcsCRLF            = gcsCR + gcsLF;

const gcsActiveChrome    = "installed-chrome.txt";
const gcsInactiveChrome  = "inactive-chrome.txt";
const gcsChromeRDF       = "chrome.rdf";
const gcsOverlayinfo     = "overlayinfo";
const gcsXUL_Fastload    = "XUL";
const gcsMnenhy          = "mnenhy";
const gcsSuffixJAR       = ".jar";
const gcsSuffixRDF       = ".rdf";
const gcsJarProtocol     = "jar:";
const gcsJarJar          = gcsSuffixJAR + "!";

const gcsChroManTree     = "mnenhyChroManTree";
const gcsChroManTreeCh   = "mnenhyChroManTreeChildren";
const gcsChroManTreeSel  = "mnenhyChroManTreeSelect";

const gcsCTKnown         = "{Known}";
const gcsCTUnknown       = "{Unknown}";
const gcsCTPackage       = "{Pack}";
const gcsCTLeaf          = "{Leaf}";
const gcsPackName        = "mnenhyPackName";
const gcsPackJar         = "mnenhyPackJar";
const gcsChromeActive    = "subscribedColumn2";
const gcsChromeActiveYes = "subscribed";
const gcsChromeJar       = "unreadButtonColHeader";
const gcsChromeJarYes    = "unread";
const gcsAttrProperties  = "properties";
const gcsAttrLabel       = "label";
const gcsAttrLine        = "mnenhyline";
const gcsAttrLevel       = "mnenhylevel";
const gcsAttrContainer   = "container";
const gcsAttrOpen        = "open";
const gcsAttrDisabled    = "disabled";
const gcsTreeCellImage   = "image";
const gcsTree            = "tree";
const gcsTreeChildren    = "treechildren";
const gcsTreeSeparator   = "treeseparator";
const gcsTreeItem        = "treeitem";
const gcsTreeRow         = "treerow";
const gcsTreeCell        = "treecell";

const gciKeySpace        = decodeURI(GetMnenhyStr("MnDialog_ChroMan_Key_Activate")).charCodeAt(0);
const gciKeyJar          = decodeURI(GetMnenhyStr("MnDialog_ChroMan_Key_Jar"     )).charCodeAt(0);


// |
// | vars
// |
var gDirServiceProp = Components.classes["@mozilla.org/file/directory_service;1"].getService(Components.interfaces.nsIProperties);
var gChromeDir      = gDirServiceProp.get("AChrom", Components.interfaces.nsIFile);

var goChromeTree    = null;
var giChromeLines   = 0;


// |
// | functions
// |
function MakeAtom(sValue)
{ // cast String to nsIAtom
  return Components.classes["@mozilla.org/atom-service;1"].getService(Components.interfaces.nsIAtomService).getAtom(sValue);
}


function DeleteChromeRDF()
{ // remove $MozillaProgramDir/chrome/chrome.rdf
  DeleteFileOrDir(gChromeDir.path, gcsChromeRDF);
}


function DeleteUserChromeRDF()
{ // remove $UserProfileDir/chrome/chrome.rdf
  var gChromeDir = gDirServiceProp.get("UChrm", Components.interfaces.nsIFile);
  DeleteFileOrDir(gChromeDir.path, gcsChromeRDF);
}


function DeleteXulFastload()
{ // remove $UserProfileDir/xul.mfl (resp. xul.mfasl)
  try
  {
    var faslService = Components.classes['@mozilla.org/fast-load-service;1'].getService(Components.interfaces.nsIFastLoadService);
    var faslFile    = faslService.newFastLoadFile(gcsXUL_Fastload);
    faslFile.remove(false);
  }
  catch(ignore)
  {
  }
}


function DeleteOverlayinfo()
{ // remove $MozillaProgramDir/chrome/overlayinfo
  DeleteFileOrDir(gChromeDir.path, gcsOverlayinfo);
}


function MnenhyPurifyChrome()
{ // remove $MozillaProgramDir/chrome/chrome.rdf and $MozillaProgramDir/chrome/overlayinfo
  DeleteChromeRDF();
  DeleteUserChromeRDF()
  DeleteOverlayinfo();
  DeleteXulFastload();
}


// remove $MozillaProgramDir/chrome/$sPackage.jar
function DeleteChromeJAR(sPackage)
{
  if (sPackage)
  {
    var sJar = sPackage + gcsSuffixJAR;
    try
    {
      DeleteFileOrDir(gChromeDir.path, sJar);
    }
    catch(errjar)
    {
      alert(GetMnenhyStr("MnErr_NoDelJar").replace("%s", sJar));
    }
  }
}


// remove $MozillaProgramDir/chrome/$sPackage directory
function DeleteChromePackDir(sPackage)
{
  if (sPackage)
  {
    try
    {
      DeleteFileOrDir(gChromeDir.path, sPackage);
    }
    catch(errloc)
    {
      alert(GetMnenhyStr("MnErr_NoDelLoc").replace("%s", sPackage));
    }
  }
}


function MnenhyUninstall()
{ // To completely remove Mnenhy we'd have to
  // - remove the entries in $MozillaProgramDir/chrome/installed-chrome.txt
  // - remove the entries in $MozillaProgramDir/chrome/chrome.rdf
  // - remove the entries in $MozillaProgramDir/chrome/overlayinfo/*/*.rdf
  // - remove $MozillaProgramDir/chrome/mnenhy.jar
  // - remove $MozillaProgramDir/defaults/profile/mnenhy.rdf
  // - remove $MozillaProfileDir/mnenhy.rdf
  // But we'll make some modifications to that and
  // - remove the entries in $MozillaProgramDir/chrome/installed-chrome.txt
  // - remove $MozillaProgramDir/chrome/chrome.rdf
  // - remove $MozillaProgramDir/chrome/overlayinfo
  // - remove $MozillaProgramDir/chrome/mnenhy.jar
  // - remove $MozillaProgramDir/defaults/profile/mnenhy.rdf
  // - keep $MozillaProfileDir/mnenhy.rdf

  if (confirm(GetMnenhyStr("MnDialog_ChroMan_UninstMnenhy")))
  { // set view mode back to normal
    SetHeaderView(cViewNormalHeaders);
    // uninstall Mnenhy and shutdown Mozilla
    DeinstallChrome(gcsMnenhy, gciPackTypeAll);
    MnenhyPurifyChrome();
    DeleteChromeJAR(gcsMnenhy);
    var oDefProfileDir = gDirServiceProp.get("ProfDefNoLoc", Components.interfaces.nsIFile);
    DeleteFileOrDir(oDefProfileDir.path, gcsMnenhy + gcsSuffixRDF);
    // shutdown to complete uninstall
    // (this will cause nsPrefWindows.js::onOK to bark with NS_ERROR_FAILURE
    //  in nsAppShellService::Quit because there are still windows open - but we can
    //  safely ignore that - they will be closed before the shutdown automatically.)
    goQuitApplication();
  }
}


function GetChromeFile(bInactive)
{ // create CFile for installed-chrome.txt or inactive-chrome.txt
  var oFile = null;
  try
  {
    var sFile = IsTrue(bInactive) ? gcsInactiveChrome : gcsActiveChrome;
    oFile = new CFile(gChromeDir.path);
    oFile.append(sFile);
  }
  catch(ignore){}
  return oFile;
}


function LoadChromeList(bInactive)
{ // load installed-chrome.txt or inactive-chrome.txt
  // and return it as a string
  var sReturn = null;
  var oFile = GetChromeFile(bInactive);
  if(oFile)
    sReturn = ReadFile(oFile);
  return sReturn;
}


// save sContent to installed-chrome.txt or inactive-chrome.txt
function SaveChromeList(sContent, bInactive)
{ // clean-up: remove all CR, keep only one LF, expand LF to CRLF
  sContent = sContent.replace(/\r/g,   "");
  sContent = sContent.replace(/(\n)+/g,gcsLF);
  sContent = sContent.replace(/\n/g,   gcsCRLF);
  sContent += gcsCRLF;
  // save
  var oFile = GetChromeFile(bInactive);
  if(oFile)
    WriteFile(oFile, sContent);
}


function DeinstallChrome(sPackageName, iPackageType)
{ // remove all iPackageType entries from installed-chrome.txt or inactive-chrome.txt
  RemoveChromeEntries(sPackageName, iPackageType);
  RemoveChromeEntries(sPackageName, iPackageType, true);
}


function RemoveChromeEntries(sPackageName, iPackageType, bInactive)
{ // remove all iPackageType entries from installed-chrome.txt or inactive-chrome.txt
  if (sPackageName)
  {
    var sChromeFile = LoadChromeList(bInactive);
    if (sChromeFile)
    {
      // iterate over all iPackageTypes
      var iStart = gciPackTypeContent;
      var iEnd   = gciPackTypeSkin;
      if (iPackageType)
        iStart = iEnd = iPackageType;
      for (var iType = iStart; iType <= iEnd; ++iType)
      { // delete the package
        var sType    = gcasPackTypesA[iType];
        var sResPath = gcsPackResPathExpr
                       .replace("%PackageName%", sPackageName)
                       .replace("%PackType%",    sType)
                       .replace("%Locale%",      "");
        var sEntry   = sType + "," + gcsPackInstall + "," + gcsPackURL + "," + sResPath;
        var re       = new RegExp(sEntry, "g");
        sChromeFile  = sChromeFile.replace(re, "");
      } // iterate over types
      // write file back to disk
      SaveChromeList(sChromeFile, bInactive);
    } // chrome file nonempty
  } // package given
}


// Remove this specific line from installed-chrome.txt or inactive-chrome.txt
function RemoveChromeEntryLine(sLine, bInactive)
{
  if (sLine)
  {
    var sChromeFile = LoadChromeList(bInactive);
    if (sChromeFile)
    { // delete line
      if (bInactive)
        sLine = "\\\[.+\\\]"+sLine;
      var re      = RegExp(sLine, "g"); // now that's ugly ;-)
      sChromeFile = sChromeFile.replace(re, "");
      // write file back to disk
      SaveChromeList(sChromeFile, bInactive);
    } // chrome file nonempty
  } // line given
}


function ChromeTreeElement(_label, _active, _jar, _TheLine, _index)
{ // constructor for chrome tree element object
  this.label                    = _label;    // leaf label
  this.active = this.baseactive = _active;   // package active state
  this.jar    = this.basejar    = _jar;      // package jarred state
  this.TheLine                  = _TheLine;  // original line from installed-chrome.txt
  this.index                    = _index;    // original line number (zero-based)
}
function ChromeTreeElementCopy(oChromeTreeElement)
{ // copy constructor for chrome tree element object
  this.label      = oChromeTreeElement.label;
  this.active     = oChromeTreeElement.active;
  this.jar        = oChromeTreeElement.jar;
  this.TheLine    = oChromeTreeElement.TheLine;
  this.index      = oChromeTreeElement.index;
  this.baseactive = oChromeTreeElement.baseactive;
  this.basejar    = oChromeTreeElement.basejar;
}


function ChromeTreeSort(a, b)
{ // sort two ChromeTreeElements via their label attribute
  var ka = a.label.toLowerCase();
  var kb = b.label.toLowerCase();
  if (ka > kb)
    return 1;
  else if (ka < kb)
    return -1;
  else // lower case equal => try case
    if (a.label > b.label)
      return 1;
    else if (a.label < b.label)
      return -1;
    else
      return 0; // giving up: equal!
}


function BuildChromeTree(asChromeTree, bInactive)
{ // add the installed or inactive packages to the hierarchical asChromeTree array
  // and return number of lines processed
  var iLines = 0;

  // prepare tree array
  if (!(gcsCTKnown in asChromeTree))
    asChromeTree[gcsCTKnown]   = []; // the root node for known lines
  if (!(gcsCTUnknown in asChromeTree))
    asChromeTree[gcsCTUnknown] = []; // the root node for unknown lines

  // add new items
  var sChrome = LoadChromeList(bInactive);
  if (sChrome)
  { // split file into single package lines
    var iLineNumber = -1; // zerobased
    var asChrome = sChrome.split(gcsLF);
    for (var iChrome = 0; iChrome < asChrome.length; ++iChrome)
    { // check if we know this kind of line
      var sPack = String(String(asChrome[iChrome]).replace(/^\s+/,"")).replace(/\s+$/,"");
      if (sPack)
      { // create or extract iLineNumber
        if (IsTrue(bInactive))
        { // use stored line number
          var sLine = sPack.slice(1,sPack.indexOf("]"));
          iLineNumber = Number(sLine);
          sPack = sPack.slice(sLine.length + 2);
        }
        else
        { // increment counter
          ++iLineNumber;
        }

        // split nonempyty line into parts
        var asParts = sPack.split(",");
        var bKnown = (asParts.length == 4);
        if (bKnown)
        { // check for known parts
          bKnown = (    (asParts[gciPackType] in gcasPackTypesB)
                     && (asParts[gciPackAction]==gcsPackInstall)
                     && (asParts[gciPackURL]   ==gcsPackURL)
                     && (asParts[gciPackPath])
                   );
          if (bKnown)
          { // parts okay so far; extract package and locale
            var sType        = asParts[gciPackType];
            var bIsLocale    = (gcasPackTypesB[sType] == gciPackTypeLocale);
            var sResPath     = gcsPackResPathExpr
                               .replace("%PackageName%", "(.+?)")
                               .replace("%PackType%", sType)
                               .replace("%Locale%", bIsLocale ? "(.+?)/" : "");
            var re           = new RegExp(sResPath, "g");
            re.exec(asParts[gciPackPath]);
            var bIsJAR       = !!RegExp.$1;
            var sPackageName = RegExp.$2;
            var sLocaleName  = RegExp.$4;
            var sDirName     = bIsLocale ? RegExp.$5 : RegExp.$4;
            bKnown = !!sPackageName;
            if (bKnown)
            { // put it into tree
              var aKnown = asChromeTree[gcsCTKnown];
              if (!(sPackageName in aKnown))
                aKnown[sPackageName] = [];
              var oPackage = aKnown[sPackageName];
              if (!(sType in oPackage))
                oPackage[sType] = [];
              var oDir = oPackage[sType];
              if (bIsLocale && sLocaleName)
              {
                if (!(sLocaleName in oDir))
                  oDir[sLocaleName] = [];
                oDir = oDir[sLocaleName];
              }
              // add data
              oDir[oDir.length] = new ChromeTreeElement(sDirName, !IsTrue(bInactive), bIsJAR, sPack, iLineNumber, sPackageName, sDirName);
            } // nonempty sPackageName
          } // part data okay
        } // 4 parts

        // special handling for unknown lines
        if (!bKnown)
        {
          var aUnknown = asChromeTree[gcsCTUnknown];
          aUnknown[aUnknown.length] = new ChromeTreeElement(sPack, !IsTrue(bInactive), null, sPack, iLineNumber);
        } // unknown

        // count only nonempty lines
        ++iLines;
      } // nonempyty line
    } // for all lines
//alert("#"+asChromeTree[gcsCTKnown]["comm"]["content"][0].label);
  } // nonempty file
  return iLines;
}


function BuildXULTree(asChromeTree)
{ // manually build and sort tree
  var oBaseTreeChildren = document.getElementById(gcsChroManTreeCh);
  // process Known packages first
  var aKnown = asChromeTree[gcsCTKnown];
  var oKnown = new ChromeTreeElement(GetMnenhyStr("MnDialog_ChroMan_KnownLine"), true);
  AddTreeItem(oBaseTreeChildren, aKnown, oKnown, gcsCTKnown);
  // then add separator
  oBaseTreeChildren.appendChild(document.createElement(gcsTreeSeparator));
  // and the unknown packages
  var aUnknown = asChromeTree[gcsCTUnknown];
  var oUnknown = new ChromeTreeElement(GetMnenhyStr("MnDialog_ChroMan_UnknownLine"), true);
  AddTreeItem(oBaseTreeChildren, aUnknown, oUnknown, gcsCTUnknown);
}


// Create new treeitem with level name sLevel from oChromeTreeElement and its subitems from aSubTree.
// Put both into tree as children of oBaseTreeChildren
// Return oChromeTreeElement.active status
function AddTreeItem(oBaseTreeChildren, aSubTree, oChromeTreeElement, sLevel)
{
  if (oChromeTreeElement.label)
  { // create treeitem
    var oTreeItem = document.createElement(gcsTreeItem);
    if (oTreeItem)
    { // create child rows
      if (aSubTree)
      { // create tree children
        var oTreeChildren = document.createElement("treechildren");
        if (oTreeChildren)
        {
          if (aSubTree.length)
          { // aSubTree is a numbered array and hence its elements are the tree's leafs!
            aSubTree.sort(ChromeTreeSort);
            for (var iPack = 0; iPack < aSubTree.length; ++iPack)
            {
              var oPack = aSubTree[iPack];
              oChromeTreeElement.active &= AddTreeItem(oTreeChildren, null, oPack, gcsCTLeaf);
            }
          }
          else
          { // construct level name
            var sSubLevelName = null;
            switch (sLevel)
            {
              case gcsCTKnown:
              case gcsCTUnknown:
                sSubLevelName = gcsCTPackage;
            }

            // aSubTree maybe an associative array, so we have to number it before sorting
            var aSubSort = [];
            for (var sSub in aSubTree)
              aSubSort[aSubSort.length] = new ChromeTreeElement(sSub, true, null, aSubTree[sSub]);
            aSubSort.sort(ChromeTreeSort);
            for (var iPackA = 0; iPackA < aSubSort.length; ++iPackA)
            {
              var oPackA = aSubSort[iPackA];
              var aPackSub = oPackA.TheLine;
              oPackA.TheLine = null;
              oChromeTreeElement.active &= AddTreeItem(oTreeChildren, aPackSub, oPackA, sSubLevelName);
            }
          }
          // add children to item
          oTreeItem.appendChild(oTreeChildren);
        } // if oTreeChildren
        // we have a subitem, so this is a (closed) container
        oTreeItem.setAttribute(gcsAttrContainer, "true");
        oTreeItem.setAttribute(gcsAttrOpen,      "false");
      } // if aSubTree
      // create item row
      if (sLevel)
      {
        oTreeItem.setAttribute(gcsAttrLevel, sLevel)
        oTreeItem.setAttribute(gcsAttrLabel, oChromeTreeElement.label)
      }
      AddTreeRow(oTreeItem, oChromeTreeElement);
      // attach treeitem to treechildren
      oBaseTreeChildren.appendChild(oTreeItem);
    }
  }
  return oChromeTreeElement.active;
}


function GetPropertiesStyle(oCell, bCheckStyle)
{ // get the oCell properties attribute according to bCheckStyle
  var bValue = null;  // not set
  if (oCell)
  { // get value to interprete
    var sValue = oCell.getAttribute(gcsAttrProperties);
    bCheckStyle = IsTrue(bCheckStyle);
    var sStyle = bCheckStyle ? gcsChromeActive : gcsChromeJar;
    if (sValue.search(new RegExp(gcsWordDelimExpr+sStyle+gcsWordDelimExpr)) >= 0)
    { // base style found
      sStyle = bCheckStyle ? gcsChromeActiveYes : gcsChromeJarYes;
      bValue = (sValue.search(new RegExp(gcsWordDelimExpr+sStyle+gcsWordDelimExpr)) >= 0);
    }
  } // have cell
  return bValue;
}


function SetPropertiesStyle(oElem, bValue, bCheckStyle, bIsJar)
{ // set the oElem properties attribute according to the bValue
  // if bCheckStyle, use checkmark, else use diamond
  // if bLocal is true, the default gcsChromeJarYes style is not set
  if (oElem)
  {
    var sStyle = oElem.getAttribute(gcsAttrProperties);
    if (bValue != null)
    {
      bCheckStyle = IsTrue(bCheckStyle);
      var sBase = bCheckStyle ? gcsChromeActive    : gcsChromeJar;
      var sYes  = bCheckStyle ? gcsChromeActiveYes : gcsChromeJarYes;
      // clear relevant styles
      if (sStyle)
        sStyle = sStyle.replace(new RegExp(gcsWordDelimExpr+sBase+gcsWordDelimExpr), "");
      if (sStyle)
        sStyle = sStyle.replace(new RegExp(gcsWordDelimExpr+sYes +gcsWordDelimExpr), "");
      if (sStyle && bCheckStyle && (bIsJar != null))
        sStyle = sStyle.replace(new RegExp(gcsWordDelimExpr+gcsChromeJarYes+gcsWordDelimExpr), "");
      if (sStyle)
        sStyle = sStyle.replace(new RegExp("\\s+"), "");
      // set styles
      sStyle += " " + sBase;
      if (bValue)
        sStyle += " " + sYes;
      if (bCheckStyle && (bIsJar != null))
      { // bIsJar can be true, false, null
        if (bIsJar)
          sStyle += " " + gcsChromeJarYes;
      }
    }
    oElem.setAttribute(gcsAttrProperties, sStyle);
  } // have cell
}


function AscendPropertiesStyleTree(oTreeItem, bCheckStyle)
{ // propagate the oTreeItems properties attribute to all parents
  // only process treeitems not on main level
  if (    oTreeItem
       && (oTreeItem.nodeName                       == gcsTreeItem)
       && (oTreeItem.parentNode.parentNode.nodeName != gcsTree)
     )
  { // get parent (treechildren) and compute a (virtual) properties attribute for it
    // by collecting every child treeitem's properties attribute (according to bCheckStyle)
    var iCheckStyle = bCheckStyle ? 0 : 1;
    var oTreeChildren = oTreeItem.parentNode;
    var aSiblings = oTreeChildren.childNodes;
    var bValue    = true;
    for (var iSib = 0; iSib < aSiblings.length; ++iSib)
    { // we don't need to care for treeseparators because they don't have treerows
      var oItemChildren = aSiblings[iSib].childNodes;
      for (var iChild = 0; iChild < oItemChildren.length; ++iChild)
      {
        var oChild = oItemChildren[iChild];
        if (oChild.nodeName == gcsTreeRow)
        {
          var oCell = oChild.childNodes[iCheckStyle];
          bValue &= GetPropertiesStyle(oCell, bCheckStyle);
        }
      } // for treeitem children
    } // for treeitems

    // write new value to parent treeitem
    var oParentTreeItem = oTreeChildren.parentNode;
    var aParentElems    = oParentTreeItem.childNodes;
    for (var iElem = 0; iElem < aParentElems.length; ++iElem)
    {
      var oElem = aParentElems[iElem];
      if (oElem.nodeName == gcsTreeRow)
      {
        var oCells = oElem.childNodes;
        var bOld = GetPropertiesStyle(oCells[iCheckStyle], bCheckStyle);
        if (bOld != bValue)
        { // set and propagate upwards
          SetPropertiesStyle(oCells[iCheckStyle], bValue, bCheckStyle);
          AscendPropertiesStyleTree(oParentTreeItem, bCheckStyle)
        } // set new value
        break;
      } // have row
    } // for treeitem elements
  } // have useful oTreeItem
}


function DescendPropertiesStyleTree(oTreeItem, bValue, bCheckStyle)
{ // set the oTreeItems properties attribute to bValue and propagate it to all children
  if (oTreeItem)
  { // walk through oTreeItem children to get current value and child list
    var oChildren = oTreeItem.childNodes;
    for (var iChild = 0; iChild < oChildren.length; ++iChild)
    {
      var oNode = oChildren[iChild];
      switch (oNode.nodeName)
      {
        case gcsTreeRow:
          { // set this treeitem value
            var oTreeCellName = oNode.childNodes[0];
            var oTreeCellJar  = oNode.childNodes[1];
            var oTreeElement  = eval(oTreeCellJar.getAttribute(gcsAttrLine));
            if (oTreeElement.TheLine)
            { // toggle active state
              oTreeElement.active = !IsTrue(oTreeElement.active);
              oTreeCellJar.setAttribute(gcsAttrLine, oTreeElement.toSource());
            }
            SetPropertiesStyle(bCheckStyle ? oTreeCellName : oTreeCellJar, bValue, true);
          }
          break;

        case gcsTreeChildren:
          { // iterate over the child treeitems
            var oItemNodes = oNode.childNodes;
            for (var iItem = 0; iItem < oItemNodes.length; ++iItem)
            {
              var oItem = oItemNodes[iItem];
              if (oItem.nodeName == gcsTreeItem)
                DescendPropertiesStyleTree(oItem, bValue, bCheckStyle);
            }
          } // walk down children
          break;
      } // switch children
    } // process oTreeItem children
  } // have oTreeItem
}


function AddTreeRow(oTreeItem, oChromeTreeElement)
{ // add row with two cells, set label and styles
  if (oTreeItem)
  {
    var oTreeRow = document.createElement(gcsTreeRow);
    if (oTreeRow)
    {
      var oTreeCellName = document.createElement(gcsTreeCell);
      var oTreeCellJar  = document.createElement(gcsTreeCell);
      if (oTreeCellName && oTreeCellJar)
      { // set label and active status
        var bSet = oChromeTreeElement.jar || (oChromeTreeElement.jar == null);
        SetPropertiesStyle(oTreeCellName, oChromeTreeElement.active, true, bSet);
        oTreeCellName.setAttribute(gcsAttrLabel, oChromeTreeElement.label);
        // set jar status and oChromeTreeElement
        SetPropertiesStyle(oTreeCellJar, oChromeTreeElement.jar, false);
        oTreeCellJar.setAttribute(gcsAttrLine, oChromeTreeElement.toSource());
        // attach cells to row
        oTreeRow.appendChild(oTreeCellName);
        oTreeRow.appendChild(oTreeCellJar);
        // attach row to item
        oTreeItem.appendChild(oTreeRow);
      }
    }
  }
}


// get the oTreeElement of this oTreeItem
function GetTreeElementOfItem(oTreeItem)
{
  var oTreeElement = null;

  if (oTreeItem && (oTreeItem.nodeName == gcsTreeItem))
  { // we need the treeitem's one and only treerow
    var oItemChildren = oTreeItem.childNodes;
    for (var iRow = 0; iRow < oItemChildren.length; ++iRow)
    { // we only have one row, so this is indangerous
      if (oItemChildren[iRow].nodeName == gcsTreeRow)
      { // get the jar cell
        oTreeElement = eval(oItemChildren[iRow].childNodes[1].getAttribute(gcsAttrLine));
        break;
      } // treerow
    } // for all treeitem childnodes
  } // if treeitem

  return oTreeElement;
}


// get the oTreeElement at index
function GetTreeElementByIndex(index)
{
  var oTreeElement = null;

  if (index >= 0)
  {
    var oTreeItem = goChromeTree.contentView.getItemAtIndex(index);
    oTreeElement  = GetTreeElementOfItem(oTreeItem);
  }

  return oTreeElement;
}


function CheckTreeSelect(event)
{ // remove, extract, purge are only allowed if something known is selected
  var bDisabled = true;
  var iIndex    = goChromeTree.currentIndex;
  if (iIndex >= 0)
  {
    var oTreeItem = goChromeTree.contentView.getItemAtIndex(iIndex);
    if (oTreeItem.nodeName == gcsTreeItem)
    {
      var sLevel    = oTreeItem.getAttribute(gcsAttrLevel);
      bDisabled = (sLevel == gcsCTUnknown);
      if (!bDisabled)
      { // check if unknown subitem is selected
        if (sLevel == gcsCTLeaf)
        {
          var oTreeElement = GetTreeElementOfItem(oTreeItem);
          bDisabled = (oTreeElement.jar == null);
        }
      }
    } // have treeitem
  }
  var oBroadcaster = document.getElementById(gcsChroManTreeSel);
  oBroadcaster.setAttribute(gcsAttrDisabled, bDisabled);
}


function CheckTreeKeyDown(event)
{
  switch (event.keyCode)
  {
    case gciKeySpace: // Space in tree!
      ProcessTreeHit(true);
      break;

    case gciKeyJar:   // Jar-Toggle
      ProcessTreeHit(false);
      break;
  }
}


function CheckTreeClick(event)
{
  if (!event.button)
  { // where did the left button hit?
    var oRow = {}, oColID = {}, oChildElt = {};
    goChromeTree.treeBoxObject.getCellAt(event.clientX, event.clientY, oRow, oColID, oChildElt);
    if (oChildElt.value == gcsTreeCellImage)
    { // we're only interested in image clicks
      switch (oColID.value)
      {
        case gcsPackName:
          ProcessTreeHit(true);
          break;

        case gcsPackJar:   // Jar-Toggle
          ProcessTreeHit(false, oRow.value);
          break;
      } // switch column
    } // image clicked
  } // left button
}


function ProcessTreeHit(bCheckmarkHit, nRow /* = null */)
{
  if (nRow == null)
    nRow = goChromeTree.currentIndex;
  if (nRow < 0)
    return;

  var oTreeItem = goChromeTree.contentView.getItemAtIndex(nRow);
  if (!oTreeItem)
    return;

  var oItemChildren = oTreeItem.childNodes;
  // we need the treeitem's one and only treerow
  var iRow = 0;
  for (; iRow < oItemChildren.length; ++iRow)
  { // we only have one row, so this is indangerous
    if (oItemChildren[iRow].nodeName == gcsTreeRow)
    { // get the two cells
      var oRowChildren  = oItemChildren[iRow].childNodes;
      var oTreeCellName = oRowChildren[0];
      var oTreeCellJar  = oRowChildren[1];
      var oTreeElement  = eval(oTreeCellJar.getAttribute(gcsAttrLine));
      if (bCheckmarkHit)
      { // toggle the checkmark
        oTreeElement.active = !IsTrue(oTreeElement.active);
        // propagate the checkmark downwards
        DescendPropertiesStyleTree(oTreeItem, oTreeElement.active, true);
        // propagate the checkmark upwards
        AscendPropertiesStyleTree(oTreeItem, true);
      }
      else
      { // toggle the diamond
        if (oTreeElement.jar != null)
        {
          oTreeElement.jar = !IsTrue(oTreeElement.jar);
          SetPropertiesStyle(oTreeCellJar,  oTreeElement.jar,    false);
          SetPropertiesStyle(oTreeCellName, oTreeElement.active, true, oTreeElement.jar);
        }
      }
      // write back object
      oTreeCellJar.setAttribute(gcsAttrLine, oTreeElement.toSource());
    } // treerow
  } // for all treeitem childnodes
}


// extract the zip file oSrcFile into the destination directory oDestDir
// extract only files matching sPattern
function MnenhyExtractZip(/* nsIFile */ oSrcFile,
                          /* nsIFile */ oDestDir,
                          /* string  */ sPattern /* = "*" */)
{
  if (!oDestDir || !oSrcFile || !oSrcFile.exists())
    return false;

  if (!sPattern)
    sPattern = "*";

  try
  { // open zip file
    var CZipReader = new Components.Constructor("@mozilla.org/libjar/zip-reader;1", "nsIZipReader", "init");
    var oZip = new CZipReader(oSrcFile);
    oZip.open();

    // create destination dir
    const clDIRECTORY   = 0x01;
    const clPermissions = 0755;

    if(!oDestDir.exists() || !oDestDir.isDirectory())
      oDestDir.create(clDIRECTORY, clPermissions);

    // extract each and every file
    var oEntry, oTrgDir;
    var nsIZipEntry  = Components.interfaces.nsIZipEntry;
    var oTarget      = new CFile(oDestDir.path); // dummy path
    var oEntries     = oZip.findEntries(sPattern);
    while(oEntries.hasMoreElements())
    {
      oEntry = oEntries.getNext().QueryInterface(nsIZipEntry);
      // create new subdirectory if necessary
      oTarget.setRelativeDescriptor(oDestDir, oEntry.name);
      oTrgDir = oTarget.parent;
      if (!oTrgDir.exists())
        oTrgDir.create(clDIRECTORY, clPermissions);
      oZip.extract(oEntry.name, oTarget);
    }
  }
  catch(error)
  {
    MnenhyError("mnenhy-pref-chroman.js::MnenhyExtract", error);
    return false;
  }

  return true;
}


// accept only treecells with our data
function TreeCellFilter(oNode)
{
  if (oNode.nodeName == gcsTreeCell)
  {
    var oTreeElement = eval(oNode.getAttribute(gcsAttrLine));
    if (oTreeElement && oTreeElement.TheLine)
      return NodeFilter.FILTER_ACCEPT;
  }
  return NodeFilter.FILTER_SKIP;
};


// accept only treecells of package items with our data
function PackageTreeCellFilter(oNode)
{
  if (    (oNode.nodeName == gcsTreeCell)
       && (oNode.parentNode.parentNode.getAttribute(gcsAttrLevel) == gcsCTPackage)
     )
  {
      return NodeFilter.FILTER_ACCEPT;
  }
  return NodeFilter.FILTER_SKIP;
};


// process all child treecells of this oNode
// get its ChromeTreeElement and call fCallback with it
function TreeCellWalker(oNode, fCallback, fAcceptNodeFilter /* = null */, oUserData /* = null */)
{
  if (!fAcceptNodeFilter)
    fAcceptNodeFilter = TreeCellFilter;

  var oTreeWalker = document.createTreeWalker(oNode,
                                              NodeFilter.SHOW_ELEMENT,
                                              fAcceptNodeFilter,
                                              true);
  if (oTreeWalker)
  { // process all relevant treecells
    var oTreeElement = null;
    for (var oNextNode = oTreeWalker.nextNode(); oNextNode; oNextNode = oTreeWalker.nextNode())
    {
      oTreeElement = eval(oNextNode.getAttribute(gcsAttrLine));
      fCallback(oTreeElement, oUserData);
    }
  }
}


// extract the single (sub)package given by oTreeElement
function MnenhyExtractSinglePackage(oTreeElement, oUserData)
{
  var sResPath     = ",(jar:)?resource:/chrome/(.+?)(\.jar!)?/(.*)";
  var re           = new RegExp(sResPath, "g");
  re.exec(oTreeElement.TheLine);
  var sPackage     = RegExp.$2;
  var sPattern     = RegExp.$4 + "*";

  // get package jar
  var oSource = new CFile(gChromeDir.path);
  oSource.append(sPackage + gcsSuffixJAR);
  // get package dir
  var oDest = new CFile(gChromeDir.path);
  oDest.append(sPackage);
  // just do it!
  SetWaitCursor();
  var bRet = MnenhyExtractZip(oSource, oDest, sPattern);
  UnsetWaitCursor();
  return bRet;
}


// extract all child treecells of the selected treeitem
function MnenhyExtractPackage()
{
  var iSel = goChromeTree.currentIndex;
  if (iSel >= 0)
  { // walk through the selected item's child treecells
    var oTreeItem = goChromeTree.contentView.getItemAtIndex(iSel);
    TreeCellWalker(oTreeItem, MnenhyExtractSinglePackage);
  }
}


// delete the local files of the single (sub)package given by oTreeElement
function MnenhyPurgeSinglePackage(oTreeElement, oUserData)
{
  var sResPath = ",(jar:)?resource:/chrome/(.+?)(\.jar!)?/(.*)";
  var re       = new RegExp(sResPath, "g");
  re.exec(oTreeElement.TheLine);
  var sPath    = RegExp.$2;
  if (RegExp.$4)
    sPath += "/" + RegExp.$4;

  // delete local path
  if (sPath)
  { // don't kill our own dumb self. BTDT :(
    var oTarget = new CFile(gChromeDir.path);
    oTarget.setRelativeDescriptor(gChromeDir, sPath);
    SetWaitCursor();
    DeleteFileOrDir(oTarget.path, "");
    UnsetWaitCursor();
  }
  return true;
}


// delete the local files of all child treecells of the selected treeitem
function MnenhyPurgePackage()
{
  var iSel = goChromeTree.currentIndex;
  if (iSel >= 0)
  { // walk through the selected item's child treecells
    var oTreeItem = goChromeTree.contentView.getItemAtIndex(iSel);
    TreeCellWalker(oTreeItem, MnenhyPurgeSinglePackage);
  }
}


// delete all children and subchildren of oNode
function MnenhyDeleteNodeTree(oNode)
{
  if (oNode)
  {
    if (oNode.hasChildNodes())
    {
      for (var iChild = 0; iChild < oNode.childNodes.length; ++iChild)
        MnenhyDeleteNodeTree(oNode.childNodes[iChild]);
    }

    if (oNode.parentNode)
    {
      oNode.parentNode.removeChild(oNode);
    }
  }
}


// remove subpackage given by oTreeElement
function MnenhyRemoveSingleSubpackage(oTreeElement, oUserData)
{
  var sResPath = ",(jar:)?resource:/chrome/(.+?)(\.jar!)?/(.*)";
  var re       = new RegExp(sResPath, "g");
  re.exec(oTreeElement.TheLine);
  var sPath    = RegExp.$2;
  if (RegExp.$4)
    sPath += "/" + RegExp.$4;

  // delete local path
  if (sPath)
  { // don't kill our own dumb self. BTDT :(
    var oTarget = new CFile(gChromeDir.path);
    oTarget.setRelativeDescriptor(gChromeDir, sPath);
    SetWaitCursor();
    DeleteFileOrDir(oTarget.path, "");  // remove local files
    RemoveChromeEntryLine(oTreeElement.TheLine);
    RemoveChromeEntryLine(oTreeElement.TheLine, true);
    UnsetWaitCursor();
  }
  return true;
}


// Delete all local files of the package, *delete* its entries from
// installed-chrome.txt (without storing them in inactive-chrome.txt!),
// delete .jar if in package treeitem, and purify chrome
// Treeview will be updated, too.
function MnenhyRemovePackage()
{
  if (MnenhyConfirm(GetMnenhyStr("MnDialog_ChroMan_Remove")))
  { // purge local files and remove entries in installed-chrome.txt and inactive-chrome.txt
    var iSel = goChromeTree.currentIndex;
    if (iSel >= 0)
    { // first iterate over the subpackages
      var oTreeItem = goChromeTree.contentView.getItemAtIndex(iSel);
      TreeCellWalker(oTreeItem, MnenhyRemoveSingleSubpackage);
    }
    // walk tree upwards to get all dead branches
    var oTreeChildren = oTreeItem.parentNode;
    while (oTreeChildren.childNodes.length < 2)
    {
      oTreeItem     = oTreeChildren.parentNode;
      oTreeChildren = oTreeItem.parentNode;
    }
    // if at package level or higher, delete local directory and jar file; clean chrome registry
    switch (oTreeItem.getAttribute(gcsAttrLevel))
    {
      case gcsCTPackage:
      case gcsCTKnown:
        {
          var sPackage = oTreeItem.getAttribute(gcsAttrLabel);
          if (sPackage)
          {
            DeleteChromePackDir(sPackage);
            DeleteChromeJAR(sPackage);
          }
        }
    }
    MnenhyPurifyChrome();
    // delete oTreeItem with all children from tree
    MnenhyDeleteNodeTree(oTreeItem)
  }
}


function InitChroManTree()
{ // load installed chromes into global chrome tree array and squash it
  goChromeTree = document.getElementById(gcsChroManTree);
  var asChromeTree = [];
  giChromeLines  = BuildChromeTree(asChromeTree);
  giChromeLines += BuildChromeTree(asChromeTree, true);
  BuildXULTree(asChromeTree);
  asChromeTree = null;

  // register pref exit functions
  parent.hPrefWindow.registerOKCallbackFunc(ExitChroManTree);
}


// collect TheLines from ChroManTree, set jat status, and part active / inactive
function MnenhyRebuildSingleChromeLine(oTreeElement, oUserData)
{
  if (oTreeElement.index != null)
  {
    if (oTreeElement.jar != null) // jar is tristate!
    { // set jar status if known line
      var asParts = oTreeElement.TheLine.split(",");
      var sPackResPathExpr = "(jar:)?(resource:/chrome/.*?)(\.jar!)?(/.*/)";
      var re = new RegExp(sPackResPathExpr, "g");
      re.exec(asParts[3]);
      oUserData.bDirty |= (oTreeElement.jar != oTreeElement.basejar);
      asParts[3] = (oTreeElement.jar ? gcsJarProtocol : "")
                 + RegExp.$2
                 + (oTreeElement.jar ? gcsJarJar      : "")
                 + RegExp.$4;
      oTreeElement.TheLine = asParts.join(",");
    }
    // prepare equal lengthed prefix
    var sLine = oUserData.sIndexBase + oTreeElement.index;
    sLine = "[" + sLine.slice(String(oTreeElement.index).length) + "]" + oTreeElement.TheLine;
    if (oTreeElement.active)
      oUserData.aActive  [oUserData.aActive.length]   = sLine;
    else
      oUserData.aInactive[oUserData.aInactive.length] = sLine;
    oUserData.bDirty |= (oTreeElement.active != oTreeElement.baseactive);
  }
}


// walk down the package tree and put package lines in installed-chrome.txt or inactive-chrome.txt
function ExitChroManTree(event)
{ // do we actually need to save something?
  var oFileContent = {bDirty:     false,
                      aActive:    [],
                      aInactive:  [],
                      sIndexBase: String(giChromeLines).replace(/./g,"0")};
  var oTreeCh = document.getElementById(gcsChroManTreeCh);
  TreeCellWalker(oTreeCh, MnenhyRebuildSingleChromeLine, null, oFileContent);

  if (oFileContent.bDirty && MnenhyConfirm(GetMnenhyStr("MnDialog_ChroMan_PrefOK")))
  { // change chrome-lists and shutdown Mozilla
    // sort the files according to their original line numbers
    oFileContent.aActive.sort();
    oFileContent.aInactive.sort();
    // transform arrays into strings
    var sActive   = oFileContent.aActive.join(gcsCRLF);
    var sInactive = oFileContent.aInactive.join(gcsCRLF);
    // remove line numbers from installed-chrome
    sActive = sActive.replace(/\[.*\]/g, "");
    // write to file
    SaveChromeList(sActive,  false);
    SaveChromeList(sInactive, true);
    MnenhyPurifyChrome();
    // shutdown to complete deactivation
    // (this will cause nsPrefWindows.js::onOK to bark with NS_ERROR_FAILURE
    //  in nsAppShellService::Quit because there are still windows open - but we can
    //  safely ignore that - they will be closed before the shutdown automatically.)
    goQuitApplication();
  }
}
