/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Mnenhy.
 *
 * The Initial Developer of the Original Code is Karsten Duesterloh.
 *
 * Portions created by the Initial Developer are Copyright (C) 2003
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *  Karsten Duesterloh <mnenhy@tprac.de>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */
// +-
// |    mnenhy.js
// |
// |    Written, forged, copied and conceived by Karsten Duesterloh <mnenhy@tprac.de>
// +-

// |
// |  global vars
// |
var gMnenhyDebug = false;    // show exceptions


// |
// |  string(bundles)
// |
var gMnenhyStrings = null;
while (!gMnenhyStrings)
{
  try
  {
    gMnenhyStrings = Components.classes['@mozilla.org/intl/stringbundle;1']
                     .getService(Components.interfaces.nsIStringBundleService)
                     .createBundle("chrome://mnenhy/locale/mnenhy.properties");
  }
  catch(error)
  {
    dump("Mnenhy::mnenhy.js: gMnenhyStrings init failed\n");
  }
}


function GetMnenhyStr(sName)
{
  var s = gMnenhyStrings.GetStringFromName(sName);
  if (!s)
    alert("Empty Property "+sName);
  return s;
}


// capitalize word starts;
// delimiters are "-,: "
function Capitalize(sText)
{
  var rs   = [];
  var s    = sText.toLowerCase();
  var bCap = true;

  for (var i = 0; i < s.length; ++i)
  {
    var c = s[i];
    switch(c)
    {
      case "-":
      case ",":
      case ":":
      case " ":
        bCap = true;
        break;

      default:
        if (bCap)
        {
          c    = c.toUpperCase();
          bCap = false;
        }
    }
    rs.push(c);
  }

  return rs.join("");
}


// |
// |  Error Message
// |
function MnenhyError(source, msg)
{
  alert(GetMnenhyStr("MnErr_Fatal")+source+"\n"+msg);
}



// |
// | javascript file include support
// |

// load a JavaScript file from the chrome dir
// a file named "filename.js" must define a global variable
//  var FILENAME_MNENHY_LOAD = true;
function MnenhyInclude(sJSFileName)
{ // create checkvarname
  var start        = sJSFileName.lastIndexOf('/') + 1;
  var end          = sJSFileName.lastIndexOf('.');
  var slice        = sJSFileName.length - end;
  var checkvarname = sJSFileName.substring(start, (sJSFileName.length - slice)).toUpperCase()+"_MNENHY_LOAD";
  if (!IsTrue(checkvarname))
  { // load file
    try
    {
      var mozIJSSubScriptLoader = Components.classes["@mozilla.org/moz/jssubscript-loader;1"].getService(Components.interfaces.mozIJSSubScriptLoader);
      mozIJSSubScriptLoader.loadSubScript("chrome://mnenhy/content/" + sJSFileName);
    }
    catch(ignore){}
    // success?
    if (!IsTrue(checkvarname))
      MnenhyError("MnenhyInclude(\""+sJSFileName+"\")",GetMnenhyStr("MnErr_NoFile"));
  }
}

// check if the variable named varname exists and is true
function IsTrue(varname)
{
  var truevalue = false;
  try
  {
    truevalue = !!eval(varname);
  }
  catch(ignore){}
  return truevalue;
}



// |
// |  RDF
// |
var CFile = new Components.Constructor("@mozilla.org/file/local;1", "nsILocalFile", "initWithPath");

// initialize the RDF support
MnenhyInclude("SourceRDF.js");

const gcsMnenhyPackage   = "mnenhy";
const gcsMnenhyNamespace = "http://mnenhy.tprac.de/rdf#";

var gMnenhyRDF = new SourceRDF(GetMnenhyRDFFile(true),
                               gcsMnenhyNamespace,
                               null,
                               gcsMnenhyPackage,
                               "Seq",
                               LoadedMnenhyRDF,
                               GetMnenhyRDFFile(false));

function LoadedMnenhyRDF(oLoadedRDF)
{
  if (!oLoadedRDF)
  {
    MnenhyError("mnenhy.js::LoadedMnenhyRDF",GetMnenhyStr("MnErr_NoRDF"));
  }
  else
  { // If we have an old mnenhy.headerlist pref, convert and delete it
    var sHeaderList  = MnenhyGetPref("mnenhy.headerlist", "char");
    if (sHeaderList.length)
    {
      alert(GetMnenhyStr("MnMsg_ConvPref"))
      var asHeaderList = sHeaderList.match(/[^,]+/g);
      var sParentCont  = "header";
      var sNewCustomID = "custom" + CustomHeaderViewCount();
      oLoadedRDF.AddContainer(sParentCont);
      oLoadedRDF.AddContainer(sNewCustomID, null, sParentCont);
      oLoadedRDF.ChangeItemValue("mnenhy.headerlist", "name", sNewCustomID, sParentCont);
      for (var iHeader in asHeaderList)
        oLoadedRDF.AddNode(asHeaderList[iHeader], sParentCont+":"+sNewCustomID);

      oLoadedRDF.FlushDataSource();

      MnenhyDeletePrefBranch("mnenhy");   // kill all mnenhy prefs
    }
  }
}


// get the Mnenhy nsIFile in the user profile dir or in the default dir
function GetMnenhyRDFFile(bUser)
{
  var oRDFFile = null;
  try
  {
    var oDirServProp = Components.classes["@mozilla.org/file/directory_service;1"].getService(Components.interfaces.nsIProperties);
    var sBaseDir     = oDirServProp.get(bUser ? "ProfD" : "ProfDefNoLoc", Components.interfaces.nsIFile);

    const gcsMnenhyRDF = "mnenhy.rdf";

    oRDFFile = new CFile(sBaseDir.path);  // dir
    oRDFFile.append(gcsMnenhyRDF);         // + name
  }
  catch(ignore)
  { /* ignored */ }

  return oRDFFile;
}


// |
// |  preferences
// |
var gMnenhyPrefs = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefBranch);

function MnenhySetPref(prefstring, preftype, value)
{
    if (value != null && value != "")
    {
        switch (preftype)
        {
            case "int":
                gMnenhyPrefs.setIntPref(prefstring, value);
                break;
            case "bool":
                gMnenhyPrefs.setBoolPref(prefstring, value);
                break;
            case "char":
                gMnenhyPrefs.setCharPref(prefstring, value);
                break;
        }
    }
    else
    {
        try
        {
            gMnenhyPrefs.clearUserPref(prefstring);
        }
        catch(ignore)
        {
            if (gMnenhyDebug)
                alert("mnenhySetPref("+prefstring+","+preftype+","+value+"):\n"+ignore)
        }
    }
}

function MnenhyGetPref(prefstring, preftype)
{ // get pref value or default for that preftype
  switch (preftype)
  {
    case "int":
      try
      {
        return gMnenhyPrefs.getIntPref(prefstring);
      }
      catch(ignore)
      { // default
        return 0;
      }
      break;

    case "bool":
      try
      {
        return gMnenhyPrefs.getBoolPref(prefstring);
      }
      catch(ignore)
      { // default
        return false;
      }
      break;

    case "char":
      try
      {
        return gMnenhyPrefs.getCharPref(prefstring);
      }
      catch(ignore)
      { // default
        return "";
      }
      break;
  }
  return null;
}


function MnenhyDeletePrefBranch(sStartingAt)
{ // delete an entire branch of prefs
  return gMnenhyPrefs.deleteBranch(sStartingAt);
}


function MnenhyGetCleanStringPref(sPref)
{
  var sPrefHeaderList = MnenhyGetPref(sPref, "char");
  sPrefHeaderList = sPrefHeaderList.replace(/\s/g,"");
  sPrefHeaderList = sPrefHeaderList.toLowerCase();
  return sPrefHeaderList;
}



// |
// |  header view mode
// |

// --- view modes
const cViewBriefDirty    =  -1; // indicates "view changed!"; GetHeaderView() *NEVER* returns this
const cViewBriefHeaders  =   0;
const cViewNormalHeaders =   1;
const cViewAllHeaders    =   2;
const cViewNoHeaders     =  99;
const cViewCustomBase    = 100;

// --- view header as
const gkiOutputText      =   0; // default if not set
const gkiOutputMail      =   1;
const gkiOutputNews      =   2;
const gkiOutputMID       =   3;
// old values; ***DEPRECATED***!
const cOutputText = gkiOutputText;
const cOutputMail = gkiOutputMail;
const cOutputNews = gkiOutputNews;
const cOutputRefs = gkiOutputMID ;


// If gMnenhyRDF.general.view.viewmode exists, getHeaderView returns its value.
// Otherwise the mail.show_headers value is returned.
function GetHeaderView()
{
    var iViewMode = gMnenhyRDF.GetItemValue("viewmode", "view", "general");
    if (iViewMode==null)
        iViewMode = mnenhyGetPref("mail.show_headers","int");
    return Number(iViewMode);
}

function SetHeaderView(mode)
{
  // set standard mozilla pref; all mnenhy modes set cViewAllHeaders in mozilla
  if (mode < cViewAllHeaders)
    MnenhySetPref("mail.show_headers", "int", mode);
  else
    MnenhySetPref("mail.show_headers", "int", cViewAllHeaders);

  // set mnenhy pref
  gMnenhyRDF.ChangeItemValue(mode, "viewmode", "view", "general");
  gMnenhyRDF.FlushDataSource();
}

function CustomHeaderViewCount()
{ // "known" is not custom
  var count = gMnenhyRDF.ContainerSize("header") - 1;
  if (count < 0)
    count = 0;
  return count;
}

function CustomViewName(index)
{
  return gMnenhyRDF.GetItemValue("name", "custom"+index, "header");
}



// |
// |  File I/O
// |

// Delete a file or directory
//  sPath         the native path to the file
//  sFileOrDir    the pathless file or directory name
function DeleteFileOrDir(sPath, sFileOrDir)
{
  var oFile = new CFile(sPath);
  oFile.append(sFileOrDir);
  if (oFile.exists())
    oFile.remove(true);
}


// read nsILocalFile content into string
function ReadFile(oFile)
{
  var sContent = null;
  if(oFile.exists())
  {
    var stream = Components.classes['@mozilla.org/network/file-input-stream;1'].createInstance(Components.interfaces.nsIFileInputStream);
    stream.init(oFile, 1, 0, false); // open as "read only"

    var scriptableStream = Components.classes['@mozilla.org/scriptableinputstream;1'].createInstance(Components.interfaces.nsIScriptableInputStream);
    scriptableStream.init(stream);

    var fileSize = scriptableStream.available();
    sContent = scriptableStream.read(fileSize);

    scriptableStream.close();
    stream.close();
  }
  return sContent;
}


// write string to disk as nsILocalFile
function WriteFile(oFile, sContent)
{
  if (oFile.exists())
    oFile.remove(true);
  oFile.create(oFile.NORMAL_FILE_TYPE, 0666);

  var stream = Components.classes['@mozilla.org/network/file-output-stream;1']
                         .createInstance(Components.interfaces.nsIFileOutputStream);
  stream.init(oFile, 2, 0x200, false); // open as "write only"
  stream.write(sContent, sContent.length);
  stream.close();
}


// |
// |  helper functions
// |
function SetWaitCursor(oWindow)
{
  var oWin = oWindow ? oWindow : window;
  oWin.setCursor("wait");
}
function UnsetWaitCursor(oWindow)
{
  var oWin = oWindow ? oWindow : window;
  oWin.setCursor("auto");
}

function MnenhyMessageBox(sContent, sTitle, sButtons, sIconClass)
{
  var oText = {value:      null, 
               sTitle:     sTitle, 
               sContent:   sContent, 
               sButtons:   sButtons,
               sIconClass: sIconClass}; // alert-icon, question-icon, message-icon, error-icon
  window.openDialog("chrome://mnenhy/content/mnenhy-dialog.xul",
                    "mnenhyConfirmDialog",
                    "chrome,modal,centerscreen",
                    oText);
  return oText.value;
}
function MnenhyAlert(sContent, sTitle /* = "Alert" */)
{
  if (!sTitle)
    sTitle = GetMnenhyStr("MnDialog_Alert");
  return MnenhyMessageBox(sContent, sTitle, "accept", "spaced alert-icon");
}
function MnenhyConfirm(sContent, sTitle /* = "Confirm" */)
{
  if (!sTitle)
    sTitle = GetMnenhyStr("MnDialog_Confirm");
  return MnenhyMessageBox(sContent, sTitle, null, "spaced question-icon");
}

function MnenhyAlertService(sContent, sTitle /* = "Alert" */)
{
  if (!sTitle)
    sTitle = GetMnenhyStr("MnDialog_Alert");
  var promptService = Components.classes["@mozilla.org/embedcomp/prompt-service;1"]
                                .getService(Components.interfaces.nsIPromptService);
  promptService.alert(window, sTitle, sContent);
}
function MnenhyConfirmService(sContent, sTitle /* = "Confirm" */)
{
  if (!sTitle)
    sTitle = GetMnenhyStr("MnDialog_Confirm");
  var promptService = Components.classes["@mozilla.org/embedcomp/prompt-service;1"]
                                .getService(Components.interfaces.nsIPromptService);
  return promptService.confirm(window, sTitle, sContent);
}

function CopyStringToClipboard(sText)
{
  var clipboard = Components.classes["@mozilla.org/widget/clipboardhelper;1"]
                            .getService(Components.interfaces.nsIClipboardHelper);
  clipboard.copyString(sText);
}

